
#include <cppunit/CompilerOutputter.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/ui/text/TestRunner.h>
#include "SWWReaderTest.h"


// Registers the fixture
CPPUNIT_TEST_SUITE_REGISTRATION( SWWReaderTest );



void SWWReaderTest::setUp()
{
    _sww = new SWWReader("../SWWReaderTest/test.sww");
}


void SWWReaderTest::testValid()
{
    CPPUNIT_ASSERT( _sww->isValid() );
}



void SWWReaderTest::testNumberOfVertices()
{
    const size_t nvertices = 20;
    CPPUNIT_ASSERT_EQUAL( _sww->getNumberOfVertices(), nvertices );
}



void SWWReaderTest::testTime()
{
    const unsigned int ntimesteps = 3;
    CPPUNIT_ASSERT_EQUAL( _sww->getNumberOfTimesteps(), ntimesteps );

    // hard-coded values from sww file
    const float time_array[3] = {0.0, 0.5, 1.0};
    for (unsigned int i=0; i<ntimesteps; i++)
        CPPUNIT_ASSERT_EQUAL( _sww->getTime(i), time_array[i] );
}



void SWWReaderTest::testBedslopeVertexArray()
{
    osg::ref_ptr<osg::Vec3Array> actual = _sww->getBedslopeVertexArray();

    // expected number of bedslope vertices
    const size_t nvertices = 20;
    CPPUNIT_ASSERT_EQUAL( actual->size(), nvertices );

    // hard-coded values extracted from sww file
    using osg::Vec3;
    using osg::Vec3Array;
    Vec3Array *expected = new Vec3Array(20);
    (*expected)[ 0] = Vec3( 0.000000, 0.000000, 0.000000 );
    (*expected)[ 1] = Vec3( 0.000000, 0.666667, 0.222222 );
    (*expected)[ 2] = Vec3( 0.000000, 1.333333, 0.444444 );
    (*expected)[ 3] = Vec3( 0.000000, 2.000000, 0.666667 );
    (*expected)[ 4] = Vec3( 0.500000, 0.000000, -0.166667 );
    (*expected)[ 5] = Vec3( 0.500000, 0.666667, 0.055556 );
    (*expected)[ 6] = Vec3( 0.500000, 1.333333, 0.277778 );
    (*expected)[ 7] = Vec3( 0.500000, 2.000000, 0.500000 );
    (*expected)[ 8] = Vec3( 1.000000, 0.000000, -0.333333 );
    (*expected)[ 9] = Vec3( 1.000000, 0.666667, -0.111111 );
    (*expected)[10] = Vec3( 1.000000, 1.333333, 0.111111 );
    (*expected)[11] = Vec3( 1.000000, 2.000000, 0.333333 );
    (*expected)[12] = Vec3( 1.500000, 0.000000, -0.500000 );
    (*expected)[13] = Vec3( 1.500000, 0.666667, -0.277778 );
    (*expected)[14] = Vec3( 1.500000, 1.333333, -0.055556 );
    (*expected)[15] = Vec3( 1.500000, 2.000000, 0.166667 );
    (*expected)[16] = Vec3( 2.000000, 0.000000, -0.666667 );
    (*expected)[17] = Vec3( 2.000000, 0.666667, -0.444444 );
    (*expected)[18] = Vec3( 2.000000, 1.333333, -0.222222 );
    (*expected)[19] = Vec3( 2.000000, 2.000000, 0.000000 );

    for (size_t i=0; i<nvertices; i++)
        CPPUNIT_ASSERT_VEC3_EQUAL( actual->at(i), expected->at(i) );
}



void SWWReaderTest::testBedslopeIndexArray()
{
    osg::ref_ptr<osg::UIntArray> actual = _sww->getBedslopeIndexArray();

    // expected number of bedslope indices
    const size_t nindices = 24*3;
    CPPUNIT_ASSERT_EQUAL( actual->size(), nindices );

    // hard-coded values extracted from sww file
    const unsigned int expected[72] = { 4, 5, 0,  1, 0, 5,  5, 6, 1, 
        2, 1, 6,  6, 7, 2,  3, 2, 7,  8, 9, 4,  5, 4, 9,  9, 10, 5, 
        6, 5, 10,  10, 11, 6,  7, 6, 11,  12, 13, 8,  9, 8, 13,
        13, 14, 9,  10, 9, 14,  14, 15, 10,  11, 10, 15,  16, 17, 12, 
        13, 12, 17,  17, 18, 13,  14, 13, 18,  18, 19, 14,  15, 14, 19 };

    for (size_t i=0; i<nindices; i++)
        CPPUNIT_ASSERT_EQUAL( actual->at(i), expected[i] );
}



void SWWReaderTest::testConnectivity()
{
    // shared vertices for triangle 17
    triangle_list actual = _sww->getConnectivity(17);

    const size_t nshared = 3;
    CPPUNIT_ASSERT_EQUAL( actual.size(), nshared );
    const unsigned int expected[nshared] = { 18, 19, 20 };
    for (size_t i=0; i<nshared; i++)
        CPPUNIT_ASSERT_EQUAL( actual.at(i), expected[i] );
}



void SWWReaderTest::testBedslopeNormalArray()
{
    osg::ref_ptr<osg::Vec3Array> actual = _sww->getBedslopeNormalArray();

    // expected number of bedslope normals
    const size_t nvertices = 24;
    CPPUNIT_ASSERT_EQUAL( actual->size(), nvertices );

    // hard-coded values (bedslope is flat plane)
    using osg::Vec3;
    using osg::Vec3Array;
    Vec3Array *expected = new Vec3Array; 
    expected->assign(24, Vec3( 0.301511, -0.301511, 0.904534 ) );

    for (size_t i=0; i<nvertices; i++)
        CPPUNIT_ASSERT_VEC3_EQUAL( actual->at(i), expected->at(i) );
}




void SWWReaderTest::tearDown()
{
}




int main(int argc, char* argv[])
{
  // Get the top level suite from the registry
  CppUnit::Test *suite = CppUnit::TestFactoryRegistry::getRegistry().makeTest();

  // Add test
  CppUnit::TextUi::TestRunner runner;
  runner.addTest( suite );

  // Change default outputter to a compiler error format outputter
  runner.setOutputter( new CppUnit::CompilerOutputter( &runner.result(), std::cerr ) );

  // Run tests
  bool wasSucessful = runner.run();

  // Error code 1 if any tests failed
  return wasSucessful ? 0 : 1;
}
